/**************************************************************************
 * Version: Simplified BSD License
 * URL: http://www.opensource.org/licenses/bsd-license.php
 *
 * Copyright (c) 2010, Joseph LaFata
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  - Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  - Neither the name of cppbench nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef BENCHRESULT_H_
#define BENCHRESULT_H_

#include "DllExport.h"

#include "Stopwatch.h"

#include <iomanip>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>

/**
 * @brief the types of time provided.
 */
namespace TimeType
{
	/**
	 * The different types of time
	 * that are contained by the BenchResult.
	 */
	enum Type
	{
		WALL_CLOCK = 0,//!< wall clock time.
		USER,          //!< user cpu time.
		SYSTEM,        //!< system or kernel cpu time.
		LAST
	};

	/// String representations of the TimeType:Type enumeration
	const char TypeStrings[LAST][100] =
		{ "wall-clock", "user", "system" };
}

/**
 * @brief the values of time provided.
 */
namespace TimeValue
{
	/**
	 * The different values related to the times
	 * that are contained by the BenchResult.
	 */
	enum Value
	{
		AVERAGE = 0,//!< average
		MINIMUM,    //!< minimum
		MAXIMUM,    //!< maximum
		STD_DEV,    //!< standard deviation
		LAST
	};

	/// String representations of the TimeValue::Value enumeration
	const char ValueStrings[LAST][100] =
		{ "average", "min", "max", "std-dev" };
}

/** @brief A Benchmark Result.
 *
 * This is a container class to store the
 * results of running a benchmark.
 */
class CPPBENCH_API BenchResult
{
public:
	BenchResult();

	/** @brief Create a BenchResult from a vector of Stopwatch::Time values
	 *
	 * This takes a vector of Stopwatch::Time values and calculates
	 * the supported statistics for the times.
	 *
	 */
	BenchResult(std::vector<Stopwatch::Time>& times);
	~BenchResult();

	void setName(const std::string& n) { name = n; }

	/** @brief Get the benchmark name.
	 *
	 * Get the name of the benchmark these results are for.
	 *
	 * @return The name of the benchmark.
	 */
	std::string getName() const { return name; }

	void setValue(const std::string& v) { val = v; }

	/** @brief Get the value.
	 *
	 * Get a string representation of the value this benchmark was run with.
	 *
	 * @return A string representation of the value.
	 */
	std::string getValue() const { return val; }

	void setTimeValue(TimeType::Type type, TimeValue::Value valueType, double value);

	/**
	 * Get the time value for the given parameters.
	 *
	 * For example, if you want the average wall clock time for this
	 * benchmark/value pair pass in TimeType::WALL_CLOCK and
	 * TimeValue::AVERAGE.
	 *
	 * @param type The type of time to get information about. (See TimeType::Type)
	 * @param valueType The value type of the information you're
	 * 		looking for. (See TimeValue::Value)
	 * @return The value for the parameters passed in.
	 */
	double getTimeValue(TimeType::Type type, TimeValue::Value valueType) const;

private:
	
#ifdef WIN32
#pragma warning( push )
#pragma warning( disable : 4251 )
#endif

    std::string name;
	std::string val;

#ifdef WIN32
#pragma warning( pop )
#endif

	double timeValues[TimeType::LAST][TimeValue::LAST];
};

#endif /* BENCHRESULT_H_ */
