/**************************************************************************
 * Version: Simplified BSD License
 * URL: http://www.opensource.org/licenses/bsd-license.php
 *
 * Copyright (c) 2010, Joseph LaFata
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  - Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  - Neither the name of cppbench nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef _BENCHMARK_H
#define _BENCHMARK_H
#pragma once

#include "BenchFormatter.h"
#include "BenchResult.h"
#include "Stopwatch.h"

#include <iomanip>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>

#include <float.h>
#include <math.h>

/** @brief The benchmark class.
 *
 * All benchmarks using this framework should
 * inherit from this base class.
 */
template<typename S>
class Benchmark
{
public:
	Benchmark(): mName("") {}
	Benchmark(std::string n): mName(n) {}
	virtual ~Benchmark() {}

    /** @brief Set the value for the benchmark.
     *
     * @param val S the value.
     *
     * This can really be anything, but should be able
     * to be taken as input to the std::stringstream left-shift
     * operator.
     */
	void setValue(S val) { mVal = val; }

	/** @brief Get the value for the current run of the benchmark
	 *
	 * @return the value
	 *
	 * The value could potentially be updated before each time the
	 * bench function is called.  This function is the correct way to
	 * get the current value to use as input to the benchmark.
	 */
	S getValue() const { return mVal; }

    /** @brief The pre benchmark task.
     *
     * @return virtual void
     *
     * Sometimes benchmarks require tasks to be done before
     * the code that you're interested in benchmarking.  This
     * is the place to take care of that.  The runtime if this
     * function will not be included in the total benchmark runtime.
     */
	virtual void preBench() {}


    /** @brief the benchmark function.
     *
     * This is the function that is timed when running the benchmark.
     */
	virtual void bench() = 0;

    /** @brief the post benchmark function.
     *
     * This should take care of any cleanup required for the benchmark
     * to be run again.  It should return the benchmark back to its
     * original state, as the same benchmark object will be used for
     * multiple runs.
     */
	virtual void postBench() {}

    /** @brief Get the name of the benchmark.
     *
     * @return std::string the name of the benchmark.
     *
     */
	std::string getName() const { return mName; }

    /** @brief Set the name of the benchmark.
     *
     * @param n std::string the name.
     */
	void setName(std::string n) { mName = n; }


    /** @brief run the benchmark.
     *
     * @param repeats The number of times to repeat the benchmark.
     * @param warmup The number of times to run the benchmark before beginning timing.
     *
     * @return BenchResult The results from running the benchmark.
     *
     * Run this benchmark the number of times given by
     * repeats.  After it's done running, it will find the
     * average runtime and the standard deviation of the
     * runtimes.  The information will be returned in a
     * BenchResult object.
     *
     */
    BenchResult run(unsigned repeats, unsigned warmup);
private:
    std::string mName;
	S mVal;
};

/** @brief A group of benchmarks to run together.
 *
 * They should generally be comparable to one another.  This is
 * partially enforced by each benchmark being called with the
 * same value.
 */
template<typename S>
class BenchGroup
{
public:
    BenchGroup() : mMaxRuntime(DBL_MAX), mRepeats(5), mWarmup(0) {}
    ~BenchGroup();

    /** @brief Add a benchmark to this benchmark group.
     *
     * @param bench
     *
     * The memory for the benchmarks added to a BenchGroup
     * is cleaned up by the BenchGroup upon destruction.
     *
     * Benchmarks are run in the order they are added to
     * this function.
     *
     */
    void addBench(Benchmark<S> *bench) { mBenches.push_back(bench); }
    void addVal(const S val) { mVals.push_back(val); }

    /** @brief The maximum runtime for the benchmarks of this group. default = unlimited
     *
     * @param d const double.
     *
     * @return void
     *
     * Some benchmarks take a long time to run.  After a certain point
     * you may not be getting any more information from running that benchmark.
     * When set if a benchmark of a given value takes longer than MaxRuntime,
     * that benchmark will not be run for future values.
     */
    void setMaxRuntime(const double d) { mMaxRuntime = d; }
    double getMaxRuntime() const { return mMaxRuntime; }

    /** @brief The number of times to repeat this benchmark. default = 5
     *
     * @param i The number of times to repeat this benchmark.
     */
    void setRepeats(const unsigned i) { mRepeats = i; }
    unsigned getRepeats() const { return mRepeats; }

    /** @brief The number of warmup runs to do. default = 0
     *
     * Sometimes it's a good idea to run a benchmark a few times
     * before starting to record results.  Warmup runs do not count towards
     * a benchmark's runtime.
     */
    void setWarmup(const unsigned i) { mWarmup = i; }
    unsigned getWarmup() const { return mWarmup; }

    /** @brief Run all benchmarks in this group with the given values.
     *
     * This will run the benchmarks in the order they were added to the
     * BenchGroup.  It will run each benchmark for the current value
     * before moving on to the next value.
     *
     * @param formatter The formatter to use to output the results.
     */
    void run(BenchFormatter& formatter) const;

protected:
    double mMaxRuntime;
    unsigned mRepeats;
    unsigned mWarmup;

    std::vector<Benchmark<S> *> mBenches;
    std::vector<S> mVals;
};

template<typename S>
BenchResult Benchmark<S>::run(unsigned repeats, unsigned warmup)
{
	for (unsigned i = 0; i < warmup; i++)
	{
		preBench();
		bench();
		postBench();
	}

    Stopwatch timer;
    std::vector<Stopwatch::Time> times;
    for (unsigned i = 0; i < repeats; i++)
    {
        preBench();

        timer.start();
        bench();
        timer.stop();

        postBench();

        times.push_back(timer.getTime());
        timer.reset();
    }

    BenchResult result(times);
    result.setName(getName());

    std::stringstream tmp;
    tmp.precision(15);
    tmp << mVal;
    result.setValue(tmp.str());

    return result;
}

template<typename S>
BenchGroup<S>::~BenchGroup()
{
    for (size_t i = 0; i < mBenches.size(); i++)
    {
        delete mBenches[i];
    }
}

template<typename S>
void BenchGroup<S>::run(BenchFormatter& formatter) const
{
    std::vector<bool> runBench;

    std::vector<std::string> values;
    std::vector<std::string> titles;
    for (size_t i = 0; i < mBenches.size(); i++)
    {
        titles.push_back(mBenches[i]->getName());
        runBench.push_back(true);
    }

    for (size_t i = 0; i < mVals.size(); i++)
    {
		std::stringstream valString;
		valString << mVals[i];
		values.push_back(valString.str());
    }
    formatter.writeTitle(titles, values);

    try
    {
        for (size_t i = 0; i < mVals.size(); i++)
        {
        	std::stringstream valString;
        	valString << mVals[i];

        	std::vector<BenchResult> results;

            for (size_t j = 0; j < mBenches.size(); j++)
            {
                if (runBench[j])
                {
                    mBenches[j]->setValue(mVals[i]);
                    results.push_back(mBenches[j]->run(mRepeats, mWarmup));

					if (results.back().getTimeValue(TimeType::WALL_CLOCK, TimeValue::AVERAGE) > mMaxRuntime)
                    {
                        runBench[j] = false;
                    }
                }
            }
            formatter.writeResult(valString.str(), results);
        }
    }
    catch(...)
    {
        // ...
    }
    formatter.finish();
}

#endif
