/**************************************************************************
 * Version: Simplified BSD License
 * URL: http://www.opensource.org/licenses/bsd-license.php
 *
 * Copyright (c) 2010, Joseph LaFata
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  - Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  - Neither the name of cppbench nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#include "Stopwatch.h"

Stopwatch::Stopwatch()
{
#ifdef WIN_TIMER
	mHandle = GetCurrentProcess();
#endif
	mWallElapsed = mUserElapsed = mSysElapsed = 0;
    mRunning = false;
}

void Stopwatch::start()
{
    mRunning = true;
#ifdef WIN_TIMER
    FILETIME creation, exit;
    QueryPerformanceCounter(&mWallMark);
    GetProcessTimes(mHandle, &creation, &exit, &mSysMark, &mUserMark);
#elif defined(POSIX_TIMER)
# ifdef MONOTONIC_TIMER
    clock_gettime(CLOCK_MONOTONIC, &mWallMark);
# else
    gettimeofday(&mWallMark, NULL);
# endif

    rusage userSys;
    if (getrusage(RUSAGE_SELF, &userSys) == 0)
    {
    	mUserMark = userSys.ru_utime;
    	mSysMark = userSys.ru_stime;
    }

#endif
}

void Stopwatch::stop()
{
    if (mRunning)
    {
#ifdef WIN_TIMER
        LARGE_INTEGER end;
		FILETIME creation, exit, sys, user;
        
		QueryPerformanceCounter(&end);
		GetProcessTimes(mHandle, &creation, &exit, &sys, &user);

        mWallElapsed += end.QuadPart - mWallMark.QuadPart;
		mUserElapsed += difftime(&mUserMark, &user);
		mSysElapsed += difftime(&mSysMark, &sys);

#elif defined(POSIX_TIMER)
# ifdef MONOTONIC_TIMER
        timespec end;
        clock_gettime(CLOCK_MONOTONIC, &end);
        mWallElapsed += ((end.tv_sec - mWallMark.tv_sec) * (uint64_t)1e9)
        		+ (end.tv_nsec - mWallMark.tv_nsec);
# else
        timeval end;
        gettimeofday(&end, NULL);
        mWallElapsed += ((end.tv_sec - mWallMark.tv_sec) * (uint64_t)1e9)
        		+ (end.tv_usec - mWallMark.tv_usec) * 1000;
# endif
        rusage userSys;
		if (getrusage(RUSAGE_SELF, &userSys) == 0)
		{
			mUserElapsed += ((userSys.ru_utime.tv_sec - mUserMark.tv_sec) * (uint64_t)1e9)
                		+ (userSys.ru_utime.tv_usec - mUserMark.tv_usec) * 1000;
			mSysElapsed += ((userSys.ru_stime.tv_sec - mSysMark.tv_sec) * (uint64_t)1e9)
			            + (userSys.ru_stime.tv_usec - mSysMark.tv_usec) * 1000;
		}
#endif
    }
    mRunning = false;
}

void Stopwatch::reset()
{
    mRunning = false;
	mWallElapsed = mUserElapsed = mSysElapsed = 0;
}

Stopwatch::Time Stopwatch::getTime()
{
    uint64_t wallElapsed = mWallElapsed;
	uint64_t userElapsed = mUserElapsed;
	uint64_t sysElapsed = mSysElapsed;

    if (mRunning)
    {
#ifdef WIN_TIMER
        LARGE_INTEGER end;
		FILETIME creation, exit, sys, user;

        QueryPerformanceCounter(&end);
		GetProcessTimes(mHandle, &creation, &exit, &sys, &user);

        wallElapsed += end.QuadPart - mWallMark.QuadPart;
		userElapsed += difftime(&mUserMark, &user);
		sysElapsed += difftime(&mSysMark, &sys);
#elif defined(POSIX_TIMER)
# ifdef MONOTONIC_TIMER
        timespec end;
		clock_gettime(CLOCK_MONOTONIC, &end);
		wallElapsed += ((end.tv_sec - mWallMark.tv_sec) * (uint64_t)1e9)
				+ (end.tv_nsec - mWallMark.tv_nsec);
# else
        timeval end;
		gettimeofday(&end, NULL);
		wallElapsed += ((end.tv_sec - mWallMark.tv_sec) * (uint64_t)1e9)
				+ (end.tv_usec - mWallMark.tv_usec) * 1000;
# endif
        rusage userSys;
		if (getrusage(RUSAGE_SELF, &userSys) == 0)
		{
			userElapsed += ((userSys.ru_utime.tv_sec - mUserMark.tv_sec) * (uint64_t)1e9)
                		+ (userSys.ru_utime.tv_usec - mUserMark.tv_usec) * 1000;
			sysElapsed += ((userSys.ru_stime.tv_sec - mSysMark.tv_sec) * (uint64_t)1e9)
			            + (userSys.ru_stime.tv_usec - mSysMark.tv_usec) * 1000;
		}
#endif
    }

	Time output;

#ifdef WIN_TIMER
    LARGE_INTEGER frequency;
	QueryPerformanceFrequency( &frequency );
	
	output.wall = ((double)wallElapsed / (double)frequency.QuadPart);
	output.user = (double)userElapsed * (double)1e-7;
	output.sys = (double)sysElapsed * (double)1e-7;

#elif defined(POSIX_TIMER)
	output.wall = (long double)wallElapsed / (long double)1e9;
	output.user = (long double)userElapsed / (long double)1e9;
	output.sys = (long double)sysElapsed / (long double)1e9;
#endif

	return output;
}

#ifdef WIN32
ULONGLONG Stopwatch::difftime(FILETIME* start, FILETIME* end)
{
    ULARGE_INTEGER st = *(ULARGE_INTEGER *)start;
    ULARGE_INTEGER et = *(ULARGE_INTEGER *)end;
    return et.QuadPart - st.QuadPart;
}
#endif
