/**************************************************************************
 * Version: Simplified BSD License
 * URL: http://www.opensource.org/licenses/bsd-license.php
 *
 * Copyright (c) 2010, Joseph LaFata
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  - Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  - Neither the name of cppbench nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef _STOPWATCH_H
#define _STOPWATCH_H
#pragma once

#include "DllExport.h"

#ifdef WIN32
# define WIN_TIMER
# include <windows.h>
#else
# include <unistd.h>
# include <sys/resource.h>
# include <sys/time.h>
# define POSIX_TIMER
# if _POSIX_TIMERS > 0 && defined(_POSIX_MONOTONIC_CLOCK)
#  define MONOTONIC_TIMER
#  include <time.h>
# endif
#endif

#if defined(__GNUC__) || (_MSC_VER >= 1600)
# include <stdint.h>
#else
typedef unsigned long long   uint64_t;
#endif

/**
 * This is a high fidelity stopwatch class that tries to choose
 * the most accurate timer for your system.
 */
class CPPBENCH_API Stopwatch
{
public:

	/**
	 * A struct containing the different time values recorded by
	 * the stopwatch.
	 */
	struct Time
	{
		double wall; //!< wall clock time
		double user; //!< user cpu time
		double sys;  //!< kernel or system cpu time
	};

	/**
	 * @brief Construct a stopwatch class.
	 */
    Stopwatch();

    /**
     * @brief Start the stopwatch.
     */
    void start();

    /**
     * @brief Stop the stopwatch.
     */
    void stop();

    /**
     * @brief Return the stopwatch to its original state.
     */
    void reset();

    /**
     * @brief Get the Time struct containing times elapsed in seconds.
     * @return The times in seconds. The Time struct contains
     * wall clock, user, and system time.
     */
    Time getTime();

private:

#ifdef WIN32
	ULONGLONG difftime(FILETIME* start, FILETIME* end);
#endif

#if defined(WIN_TIMER)
	HANDLE mHandle;
	FILETIME mUserMark;
	FILETIME mSysMark;
	LARGE_INTEGER mWallMark;
#elif defined(POSIX_TIMER)
# if defined(MONOTONIC_TIMER)
	timespec mWallMark;
# else
	timeval mWallMark;
# endif
	timeval mUserMark;
	timeval mSysMark;
#endif

	/// Elapsed times are in nanoseconds
	uint64_t mWallElapsed;
	uint64_t mUserElapsed;
	uint64_t mSysElapsed;

	bool mRunning;
};

#endif /* _STOPWATCH_H */
